#include <windows.h>
#include <direct.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include "gain_analysis.h"

#define APPNAME     "wapet"
#define VERSION     "0.3"

// -------------------------------------

void usage ()
{
    printf (
            APPNAME " version " VERSION "\n"
            "\n"
            "Usage: " APPNAME " <target> [tagging options] <encoder> <encoder options>\n"
            "\n"
            "target          : name of the file <encoder> creates\n"
            "tagging options : -t \"tag=value\" ; for example -t \"Artist=%%a\"\n"
            "                : -ape1 ; use APEv1 instead of APEv2\n"
            "encoder         : name of the encoder\n"
            "encoder options : required parameters for encoder\n"
           );
}

struct APETagFooterStruct {
    unsigned char   ID       [8];
    unsigned char   Version  [4];
    unsigned char   Length   [4];
    unsigned char   TagCount [4];
    unsigned char   Flags    [4];
    unsigned char   Reserved [8];
};

unsigned long Read_LE_Uint32 ( const unsigned char* p )
{
    return ((unsigned long)p[0] <<  0) |
           ((unsigned long)p[1] <<  8) |
           ((unsigned long)p[2] << 16) |
           ((unsigned long)p[3] << 24);
}

void Write_LE_Uint32 ( unsigned char* p, const unsigned long value )
{
    p[0] = (unsigned char) (value >>  0);
    p[1] = (unsigned char) (value >>  8);
    p[2] = (unsigned char) (value >> 16);
    p[3] = (unsigned char) (value >> 24);
}

int OpenWavFile ( FILE* in, unsigned int *samplerate, unsigned char *channels, unsigned char *format, unsigned int *datalen)
{
	char rbuffer[4];
	unsigned char ibuffer[4];
	unsigned int x;
	unsigned short bitspersample;
	unsigned int filelen, position = 0;
	fseek(in, 0, 2);
	filelen = ftell(in);
	fseek(in, 0, 0);
	fread(rbuffer, 1, 4, in);
	if ((rbuffer[0]!='R')||(rbuffer[1]!='I')||(rbuffer[2]!='F')||(rbuffer[3]!='F')) return -2;
	fseek(in,8,0);
	fread(rbuffer, 1, 4, in);
	if ((rbuffer[0]!='W')||(rbuffer[1]!='A')||(rbuffer[2]!='V')||(rbuffer[3]!='E')) return -2;
	position+=12;
	fread(rbuffer, 1, 4, in);
	if ((rbuffer[0]!='f')||(rbuffer[1]!='m')||(rbuffer[2]!='t')) return -2;
	fread(ibuffer,1,4,in);
	x = ((ibuffer[3]<<24)|(ibuffer[2]<<16)|(ibuffer[1]<<8)|(ibuffer[0]));
	if (x != 16) return -3;
	fread(ibuffer,1,2,in);
	*format = ((ibuffer[1]<<8)|(ibuffer[0]));
	if ((*format!=1)&&(*format!=3)) return -3;
	fread(ibuffer,1,2,in);
	*channels = ((ibuffer[1]<<8)|(ibuffer[0]));
	if (*channels > 2) return -3;
	fread(ibuffer,1,4,in);
	*samplerate = ((ibuffer[3]<<24)|(ibuffer[2]<<16)|(ibuffer[1]<<8)|(ibuffer[0]));
	fseek(in,6,1);
	fread(ibuffer,1,2,in);
	bitspersample = ((ibuffer[1]<<8)|(ibuffer[0]));
//	if ((*format==1)&&(bitspersample==8)) *format = 0;
/*	else*/ if ((*format==1)&&(bitspersample==16)) *format = 1;
//	else if ((*format==1)&&(bitspersample==32)) *format = 2;
//	else if ((*format==3)&&(bitspersample==32)) *format = 3;
//	else if ((*format==3)&&(bitspersample==64)) *format = 4;
	else return -3;
	position+=22;
	while (position < filelen)
	{
		fread(rbuffer, 1, 4, in);
		position+=4;
		if ((rbuffer[0]!='d')||(rbuffer[1]!='a')||(rbuffer[2]!='t')||(rbuffer[3]!='a'))
		{
			fread(ibuffer,1,4,in);
			x = ((ibuffer[3]<<24)|(ibuffer[2]<<16)|(ibuffer[1]<<8)|(ibuffer[0]));
			position+=4;
			position+=x;
			fseek(in,x,1);
		}
		else break;
	}
	fread(ibuffer,1,4,in);
	*datalen = ((ibuffer[3]<<24)|(ibuffer[2]<<16)|(ibuffer[1]<<8)|(ibuffer[0]));
	return 0;
}

int GetSamples (FILE* in, unsigned char channels, unsigned char format, unsigned int *datalen, unsigned int samples, double *left, double *right )
{
	unsigned int qw, qe;
	short *buffer;
	qe = samples*channels;
	buffer = malloc(qe*2);
	if(*datalen<(qe*2)) qe = (*datalen)/2;
	qw=fread(buffer,2,qe,in);
	(*datalen)-=(qw*2);
	if (qw != qe) *datalen = 0;
	if (channels == 1)
	{
		for(qe=0;qe<qw;qe++)
		{
			left[qe] = buffer[qe];
		}
	}
	if (channels == 2)
	{

		for(qe=0;qe<(qw/2);qe++)
		{
			left[qe] = buffer[(qe*2)];
			right[qe] = buffer[(qe*2)+1];
		}
	}
	free (buffer);
	return (qw/channels);
}

// Convert UNICODE to UTF-8
// Return number of bytes written
int unicodeToUtf8 ( const WCHAR* lpWideCharStr, char* lpMultiByteStr, int cwcChars )
{
    const unsigned short*   pwc = (unsigned short *)lpWideCharStr;
    unsigned char*          pmb = (unsigned char  *)lpMultiByteStr;
    const unsigned short*   pwce;
    size_t  cBytes = 0;

    if ( cwcChars >= 0 ) {
        pwce = pwc + cwcChars;
    } else {
        pwce = NULL;
    }

    while ( (pwce == NULL) || (pwc < pwce) ) {
        unsigned short  wc = *pwc++;

        if ( wc < 0x00000080 ) {
            *pmb++ = (char)wc;
            cBytes++;
        } else
        if ( wc < 0x00000800 ) {
            *pmb++ = (char)(0xC0 | ((wc >>  6) & 0x1F));
            cBytes++;
            *pmb++ = (char)(0x80 |  (wc        & 0x3F));
            cBytes++;
        } else
        if ( wc < 0x00010000 ) {
            *pmb++ = (char)(0xE0 | ((wc >> 12) & 0x0F));
            cBytes++;
            *pmb++ = (char)(0x80 | ((wc >>  6) & 0x3F));
            cBytes++;
            *pmb++ = (char)(0x80 |  (wc        & 0x3F));
            cBytes++;
        }
        if ( wc == L'\0' )
            return cBytes;
    }

    return cBytes;
}

int ansi_to_utf8 ( const char *ansi, char *utf8, int alen )
{
    WCHAR   *wszValue;
    int     len;

    if ( alen <= 0 ) alen = strlen ( ansi );
    if ( alen == 0 ) {
        *utf8 = '\0';
        return 0;
    }

    if ( (wszValue = (WCHAR *)malloc ((alen + 1) * 2)) == NULL ) {
        fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
        return 1;
    }

    // Convert ANSI value to Unicode
    if ( (len = MultiByteToWideChar (CP_ACP, 0, ansi, alen, wszValue, (alen + 1) * 2)) == 0 ) {
        fprintf ( stderr, APPNAME ": MultiByteToWideChar failed...\n" );
        return 1;
    }
    wszValue[len++] = L'\0';
    // Convert Unicode value to UTF-8
    if ( (len = unicodeToUtf8 (wszValue, utf8, len)) == 0 ) {
        fprintf ( stderr, APPNAME ": unicodeToUtf8 failed...\n" );
        return 1;
    }

    return 0;
}

int len_utf8 ( const char *ansi, int alen )
{
    WCHAR   *wszValue;
    char    *uszValue;
    int     len;

    if ( alen <= 0 ) alen = strlen ( ansi );
    if ( alen == 0 ) return 0;

    if ( (wszValue = (WCHAR *)malloc ((alen + 1) * 2)) == NULL ) {
        fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
        return 1;
    }
    if ( (uszValue = (char *) malloc ((alen + 1) * 3)) == NULL ) {
        fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
        return 1;
    }

    // Convert ANSI value to Unicode
    if ( (len = MultiByteToWideChar (CP_ACP, 0, ansi, alen, wszValue, (alen + 1) * 2)) == 0 ) {
        fprintf ( stderr, APPNAME ": MultiByteToWideChar failed...\n" );
        return 1;
    }
    wszValue[len++] = L'\0';
    // Convert Unicode value to UTF-8
    if ( (len = unicodeToUtf8 (wszValue, uszValue, len)) == 0 ) {
        fprintf ( stderr, APPNAME ": unicodeToUtf8 failed...\n" );
        return 1;
    }

    free ( uszValue );

    return len-1;
}

int proper_tag ( const char *tag )
{
    char *p;
    if ( (!tag) || (*tag == '\0') || (*tag == '=') ) return 0;
    if ( (p = strchr (tag, '=')) == NULL ) return 0;
    if ( *(p+1) == '\0' ) return 0;
    if ( !lstrcmpi (tag, "year=-1") ) return 0;
    return 1;
}

// Writes APEv1/v2 tag
int WriteAPETag ( FILE *fp, char **tags, size_t tag_count, int ape1 )
{
    unsigned char               temp[4];
    unsigned char               *buff;
    unsigned char               *p;
    struct APETagFooterStruct   T;
    unsigned int                flags;
    char                        *value;
    size_t                      itemlen;
    size_t                      valuelen;
    size_t                      TagCount;
    size_t                      TagSize;
    size_t                      i;

    if ( fseek ( fp, 0L, SEEK_END ) != 0 )
        return 1;

    TagCount = 0;
    TagSize = sizeof (T);                                               // calculate size of buffer needed

    for ( i = 0; i < tag_count; i++ ) {
        if ( proper_tag (tags[i]) ) {
            char *t = strchr ( tags[i], '=' ) + 1;
            itemlen = t - tags[i] - 1;
            if ( ape1 ) {
                valuelen = strlen ( t ) + 1;
            } else {
                valuelen = len_utf8 ( t, 0 );
            }
            TagCount++;
            TagSize += 8 + itemlen + 1 + valuelen;
        }
    }

    if ( TagCount == 0 ) return 0;

    if ( (buff = malloc (TagSize + sizeof (T))) == NULL ) {
        fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
        return 1;
    }

    p = buff;

    if ( !ape1 ) {
        flags  = 1<<31;                                     // contains header
        flags |= 1<<29;                                     // this is the header
        memcpy ( T.ID, "APETAGEX", sizeof (T.ID) );         // ID String
        Write_LE_Uint32 ( T.Version , 2000     );           // Version 2.000
        Write_LE_Uint32 ( T.Length  , TagSize  );           // Tag size
        Write_LE_Uint32 ( T.TagCount, TagCount );           // Number of fields
        Write_LE_Uint32 ( T.Flags   , flags    );           // Flags
        memset ( T.Reserved, 0, sizeof (T.Reserved) );      // Reserved
        memcpy ( p, &T, sizeof (T) );   p += sizeof (T);                        // insert header
    }

    for ( i = 0; i < tag_count; i++ ) {
        if ( proper_tag (tags[i]) ) {
            char *t = strchr ( tags[i], '=' ) + 1;
            itemlen = t - tags[i] - 1;
            if ( ape1 ) {
                valuelen = strlen ( t ) + 1;
                value = (char *)malloc ( valuelen+1 );
                if ( value == NULL ) {
                    fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
                    free ( buff );
                    return 1;
                }
                strcpy ( value, t );
            } else {
                valuelen = len_utf8 ( t, 0 );
                value = (char *)malloc ( valuelen+1 );
                if ( value == NULL ) {
                    fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
                    free ( buff );
                    return 1;
                }
                if ( ansi_to_utf8 (t, value, 0) != 0 ) {
                    free ( buff );
                    return 1;
                }
            }

            Write_LE_Uint32 ( temp, valuelen );
            memcpy ( p, temp, 4 );  p += 4;
            Write_LE_Uint32 ( temp, 0 );
            memcpy ( p, temp, 4 );  p += 4;

            memcpy ( p, tags[i], itemlen );
            if ( *p >= 'a' && *p <= 'z' ) *p -= 'a' - 'A';
            p += itemlen;
            *p++ = '\0';

            memcpy ( p, value, valuelen );
            p += valuelen;
            free ( value );
        }
    }

    if ( ape1 ) flags  = 0; else flags = 1<<31;         // contains header
    memcpy ( T.ID, "APETAGEX", sizeof (T.ID) );         // ID String
    if ( ape1 )
        Write_LE_Uint32 ( T.Version, 1000 );            // Version 1.000
    else
        Write_LE_Uint32 ( T.Version, 2000 );            // Version 2.000
    Write_LE_Uint32 ( T.Length  , TagSize  );           // Tag size - header
    Write_LE_Uint32 ( T.TagCount, TagCount );           // Number of fields
    Write_LE_Uint32 ( T.Flags   , flags    );           // Flags
    memset ( T.Reserved, 0, sizeof (T.Reserved) );      // Reserved
    memcpy ( p, &T, sizeof (T) );                                           // insert footer

    if ( ape1 ) i = 0; else i = sizeof (T);
    if ( fwrite (buff, 1, TagSize + i, fp) != TagSize + i ) {
        free ( buff );
        return 1;
    }

    free ( buff );

    return 0;
}

int write_tag ( const char *filename, char **tags, size_t tag_count, int ape1, int repg, double titlegain, double max_amp )
{
    FILE *fp;
	char *bufa, *bufb;
    int  error, qw;
    if ( tag_count == 0 ) return 0;

    if ( (fp = fopen (filename, "rb+")) == NULL ) {
        fprintf ( stderr, APPNAME ": failed to open file...\n" );
        return 1;
    }
	if (repg == 1)
	{
		qw=0;
		bufa = malloc(20);
		bufb = malloc(20);
		gcvt ( titlegain, 3, bufa );
		gcvt ( max_amp, 6, bufb );
		tags[tag_count] = malloc(40);
		tags[tag_count][0] = 'r';
		tags[tag_count][1] = 'e';
		tags[tag_count][2] = 'p';
		tags[tag_count][3] = 'l';
		tags[tag_count][4] = 'a';
		tags[tag_count][5] = 'y';
		tags[tag_count][6] = 'g';
		tags[tag_count][7] = 'a';
		tags[tag_count][8] = 'i';
		tags[tag_count][9] = 'n';
		tags[tag_count][10] = '_';
		tags[tag_count][11] = 't';
		tags[tag_count][12] = 'r';
		tags[tag_count][13] = 'a';
		tags[tag_count][14] = 'c';
		tags[tag_count][15] = 'k';
		tags[tag_count][16] = '_';
		tags[tag_count][17] = 'g';
		tags[tag_count][18] = 'a';
		tags[tag_count][19] = 'i';
		tags[tag_count][20] = 'n';
		tags[tag_count][21] = '=';
		qw=22;
		if (bufa[0] != '-')
		{
			tags[tag_count][qw++] = '+';
			tags[tag_count][qw++] = bufa[0];
			tags[tag_count][qw++] = bufa[1];
			tags[tag_count][qw++] = bufa[2];
			tags[tag_count][qw++] = bufa[3];
		}
		else
		{
			tags[tag_count][qw++] = bufa[0];
			tags[tag_count][qw++] = bufa[1];
			tags[tag_count][qw++] = bufa[2];
			tags[tag_count][qw++] = bufa[3];
			tags[tag_count][qw++] = bufa[4];
		}
		tags[tag_count][qw++] = ' ';
		tags[tag_count][qw++] = 'd';
		tags[tag_count][qw++] = 'B';
		tags[tag_count][qw++] = 0;
		tag_count++;
		tags[tag_count] = malloc(40);
		tags[tag_count][0] = 'r';
		tags[tag_count][1] = 'e';
		tags[tag_count][2] = 'p';
		tags[tag_count][3] = 'l';
		tags[tag_count][4] = 'a';
		tags[tag_count][5] = 'y';
		tags[tag_count][6] = 'g';
		tags[tag_count][7] = 'a';
		tags[tag_count][8] = 'i';
		tags[tag_count][9] = 'n';
		tags[tag_count][10] = '_';
		tags[tag_count][11] = 't';
		tags[tag_count][12] = 'r';
		tags[tag_count][13] = 'a';
		tags[tag_count][14] = 'c';
		tags[tag_count][15] = 'k';
		tags[tag_count][16] = '_';
		tags[tag_count][17] = 'p';
		tags[tag_count][18] = 'e';
		tags[tag_count][19] = 'a';
		tags[tag_count][20] = 'k';
		tags[tag_count][21] = '=';
		tags[tag_count][22] = bufb[0];
		tags[tag_count][23] = bufb[1];
		tags[tag_count][24] = bufb[2];
		tags[tag_count][25] = bufb[3];
		tags[tag_count][26] = bufb[4];
		tags[tag_count][27] = bufb[5];
		tags[tag_count][28] = bufb[6];
		tags[tag_count][29] = bufb[7];
		tags[tag_count][30] = 0;
		tag_count++;
		free(bufa);
		free(bufb);
	}
    if ( (error = WriteAPETag (fp, tags, tag_count, ape1)) != 0 ) {
        fprintf ( stderr, APPNAME ": tag writing failed...\n" );
    }

    fclose (fp);

    return error;
}

int main ( int argc, char **argv )
{
    // proggie <encoded file> [tagging options] <encoder + options>
    //    0          1               2..n               n+1

    char **option = NULL;
    char *cmd = NULL;
    int arg_enc = 0;
    int opt_c = 0;
    int enc_c = 0;
    int ape1 = 0;
	int repg = 0;
    int i;
	FILE *infile;
	unsigned int samplerate;
	unsigned char channels;
	unsigned char format; //0=PCM 8Bit int; 1=PCM 16Bit int; 2=PCM 32Bit int; 3=PCM 32Bit float; 4=PCM 64Bit float
	unsigned int datalen;
	double max_amp = 0;
	double titlegain;
	int qw, qe;
	double *left;
	double *right;
    if ( argc < 3 ) {
        usage ();
        return 1;
    }

    // allocate memory for maximum possible number of tagging options
    option = (char **)malloc ( sizeof(char *) * (argc-2) );
    if ( option == NULL ) {
        fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
        return 1;
    }

    for ( i = 2; i < argc; i++ ) {
        if ( !lstrcmp (argv[i], "-t") ) {  // tagging option
            if ( (++i)+1 >= argc ) {
                fprintf ( stderr, APPNAME ": no encoding settings specified...\n" );
                free ( option );
                return 1;
            }

            option[opt_c++] = argv[i];
        }
		else if ( !lstrcmp (argv[i], "-rg") ) {  // do replaygain scan
            if ( (++i)+1 >= argc ) {
                fprintf ( stderr, APPNAME ": no encoding settings specified...\n" );
                free ( option );
                return 1;
            }
			infile = NULL;
			infile = fopen(argv[i], "rb");
			if (infile == NULL)
			{
                fprintf ( stderr, APPNAME ": could not open replaygain wav-file...\n" );
                free ( option );
                return 1;
			}
			if (OpenWavFile(infile, &samplerate, &channels, &format, &datalen)!=0)
			{
                fclose(infile);
				fprintf ( stderr, APPNAME ": could not open replaygain wav-file...\n" );
                free ( option );
                return 1;
			}
			printf("Replaygain analysis... ");
			InitGainAnalysis ( samplerate );
			left = malloc(80000);
			right = malloc(80000);
			while(datalen!=0)
			{
				qw = GetSamples(infile, channels, format, &datalen, 10000, left, right);
				AnalyzeSamples(left,right,qw,channels);
				for (qe=0; qe<qw; qe++)
				{
					if (max_amp < left[qe]) max_amp = left[qe];
				}
				if (channels == 2) for (qe=0; qe<qw; qe++)
				{
					if (max_amp < right[qe]) max_amp = right[qe];
				}
			}
			max_amp/=32768.0f;
			titlegain = GetTitleGain();
			if (max_amp < 0) max_amp*=-1;
			repg=1;
			fclose(infile);
			printf("finished. \n\n");
			free(left);
			free(right);
		}
        else if ( !lstrcmp (argv[i], "-ape1") ) {  // use APEv1
            ape1 = 1;
        }
        else {
            arg_enc = i;
            enc_c = argc - arg_enc;
            break;
        }
    }
    if ( enc_c == 0 ) {
        free ( option );
        fprintf ( stderr, APPNAME ": no encoding settings specified...\n" );
        return 1;
    }

    {
        char *p;
        int cmd_len = _MAX_PATH;

        for ( i = 1; i < enc_c; i++ ) {
            cmd_len += strlen ( argv[arg_enc+i] ) + 3;
        }

        cmd = (char *)malloc ( cmd_len );
        if ( cmd == NULL ) {
            free ( option );
            fprintf ( stderr, APPNAME ": failed to allocate memory...\n" );
            return 1;
        }

        p = cmd;

        i = GetShortPathName ( argv[arg_enc], p, _MAX_PATH );
        if ( (i == 0) || (i > _MAX_PATH) ) {
            p += sprintf ( p, "%s", argv[arg_enc] );
        } else {
            p += i;
        }

        for ( i = 1; i < enc_c; i++ ) {
            p += sprintf ( p, " \"%s\"", argv[arg_enc+i] );
        }
    }
    // encode
    system ( cmd );

    i = write_tag ( argv[1], option, opt_c, ape1, repg, titlegain, max_amp);

    free ( option );
    free ( cmd );

    return i;
}
